"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeHTMLReport = void 0;
const path = require("path");
const fsa = require("fs-extra");
const child_process_1 = require("child_process");
const state_1 = require("../../state");
const pageStyles = require("./page.pc");
const styles = require("./report.pc");
const utils_1 = require("../../utils");
const html_entities_1 = require("html-entities");
const entities = new html_entities_1.Html5Entities();
const DEFAULT_OUTPUT_DIR = ".paperclip/diff/report";
const MAIN_CSS_PATH = path.join(__dirname, "../../assets/main.css");
const writeHTMLReport = (info) => __awaiter(void 0, void 0, void 0, function* () {
    const files = generateHTMLReport(info);
    const baseDir = getBaseDir(info);
    (0, utils_1.logInfo)(`Writing HTML report...`);
    const otherScreenshotIds = (0, state_1.getProjectScreenshots)(info.deltaVersion, info.manifest).map(diff => diff.id);
    for (const screenshot of (0, state_1.getProjectScreenshots)(info.currentVersion, info.manifest)) {
        fsa.cpSync((0, state_1.getScreenshotPath)(info.gitDir, screenshot.id), path.join(baseDir, screenshot.id + ".png"));
        const diff = (0, state_1.getScreenshotDiffs)(screenshot, info.manifest).find(diff => otherScreenshotIds.includes(diff.otherScreenshotId));
        if (diff) {
            fsa.cpSync((0, state_1.getScreenshotPath)(info.gitDir, diff.id), path.join(baseDir, diff.id + ".png"));
            fsa.cpSync((0, state_1.getScreenshotPath)(info.gitDir, diff.otherScreenshotId), path.join(baseDir, diff.otherScreenshotId + ".png"));
        }
    }
    for (const fileName in files) {
        const filePath = path.join(baseDir, fileName);
        fsa.writeFileSync(filePath, files[fileName]);
    }
    (0, utils_1.logInfo)(`HTML report written to ./${path.relative(info.cwd, baseDir)}/index.html`);
    if (info.open) {
        (0, child_process_1.exec)(`open ${baseDir}/index.html`);
    }
});
exports.writeHTMLReport = writeHTMLReport;
const generateHTMLReport = (info) => {
    const files = {};
    const frames = (0, state_1.getProjectFrames)((0, state_1.getProjectSnapshot)(info.currentVersion, info.manifest), info.manifest);
    for (const frame of frames) {
        files[frame.id + ".html"] = generateSnapshotPage(frame, info);
    }
    const sidebar = generateSidebarProps(info);
    // TODO - need to create summary page. For now we use alias to first page
    if (sidebar.length) {
        files["index.html"] = files[sidebar[0].id + ".html"];
    }
    files["main.css"] = fsa.readFileSync(MAIN_CSS_PATH, "utf-8");
    return files;
};
const getBaseDir = (info) => path.join(info.cwd, info.output || DEFAULT_OUTPUT_DIR, info.currentVersion, info.deltaVersion);
const generateSnapshotPage = (snapshot, info) => {
    const screenshots = (0, state_1.getFrameScreenshots)(snapshot, info.manifest);
    const sourceFile = entities.encode(snapshot.sourceFilePath);
    const filePath = path.join(info.gitDir, snapshot.sourceFilePath);
    return pageStyles.default({
        head: `<link rel="stylesheet" href="./main.css" />`,
        children: styles.Report({
            title: entities.encode(snapshot.title),
            subtitle: `<a href="vscode://file/${filePath}/">${sourceFile}</a>`,
            sidebar: generateSidebar(snapshot, info),
            content: screenshots
                .map(screenshot => {
                return generateComparison(screenshot, info);
            })
                .join("\n")
        })
    });
};
const generateComparison = (screenshot, info) => {
    const otherScreenshotIds = (0, state_1.getProjectScreenshots)(info.deltaVersion, info.manifest).map(diff => diff.id);
    const diff = (0, state_1.getScreenshotDiffs)(screenshot, info.manifest).find(diff => otherScreenshotIds.includes(diff.otherScreenshotId));
    return [
        styles.Comparison({
            children: [
                styles.Screenshot({
                    a: true,
                    title: `A: ${info.deltaVersion}`,
                    children: diff
                        ? `<img src="./${diff.otherScreenshotId}.png" />`
                        : `No screenshot found`
                }),
                styles.Screenshot({
                    b: true,
                    title: `B: ${info.currentVersion}`,
                    children: `<img src="./${screenshot.id}.png" />`
                })
            ]
        }),
        diff &&
            styles.Screenshot({
                c: true,
                title: "Changes",
                children: `<img src="./${diff.id}.png" />`
            }),
        styles.Divider({})
    ].join("\n");
};
const generateSidebarProps = (info, activeSnapshot) => {
    const project = (0, state_1.getProjectSnapshot)(info.currentVersion, info.manifest);
    const frames = (0, state_1.getProjectFrames)(project, info.manifest);
    const allProps = frames.map(frame => {
        const screenshots = (0, state_1.getFrameScreenshots)(frame, info.manifest);
        const diffs = screenshots.reduce((diffs, screenshot) => {
            return [...diffs, ...(0, state_1.getScreenshotDiffs)(screenshot, info.manifest)];
        }, []);
        const hasChanges = diffs.some(diff => diff.changeCount > 0);
        return {
            href: "./" + frame.id + ".html" + "#" + frame.id,
            id: frame.id,
            title: entities.encode(frame.title),
            src: "./" + screenshots[0].id + ".png",
            hasChanges,
            active: frame.id === (activeSnapshot === null || activeSnapshot === void 0 ? void 0 : activeSnapshot.id),
            noChanges: !hasChanges
        };
    });
    allProps.sort((a, b) => {
        return a.hasChanges ? -1 : 1;
    });
    return allProps;
};
const generateSidebar = (activeSnapshot, info) => {
    const allProps = generateSidebarProps(info, activeSnapshot);
    return allProps.map(props => styles.Thumbnail(props));
};
